#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Code to extract discussions from corrected XML files
#%%
# to work with atom
%load_ext autoreload
%autoreload 2

import pickle
import time
import xml.etree.ElementTree as ET

import sys
sys.path.append('src/python/')

import def_classes as defc
import utils_proc
import utils_annot

import os

from utils_proc import call_with_out

#%%
# specify input and output files

# needed for running in atom, can be ignored
year = '1891'
input_lastnames = "data/politicians/lastnames/" + year + "_lastnames.pickle"
input_correctedxml = "data/AB/" + year + "/04_correctedxml.tar.gz"
input_correctedmeta = "data/AB/" + year + "/03_correctedmeta.tar.gz"
output_annotatedxml = "data/AB/" + year + "/05_annotatedxml.tar.gz"

#%%
# detect arguments
input_lastnames = sys.argv[1]
input_correctedxml = sys.argv[2]
input_correctedmeta = sys.argv[3]
output_annotatedxml = sys.argv[4]

#%%
# extract suffixes, year, folder_database
suffix_tar_correctedxml = input_correctedxml.split('/')[-1].split('.tar.gz')[0]
#suffix_tar_correctedmeta = input_correctedmeta.split('/')[-1].split('.tar.gz')[0]
year = input_correctedxml.split('/')[-2]
folder_database = input_correctedxml.split(year)[0]
suffix_correctedmeta = '_metacorr'
#suffix_correctedxml = '_datacorr'
# needed to instantiate object for every document
input_rawmeta = folder_database + '/' + year + '/' + '01_rawmeta.tar.gz'

#%%
# git lfs pull necessary data
for lfsfile in [input_correctedxml, input_correctedmeta, input_rawmeta]:
    command = 'git lfs pull -I ' + lfsfile
    #print(command)
    call_with_out(command)

#%%
# TODO: exclude overlaps --> after annotation


#%%
start_time_discussions = time.time()
print('start to identify discussions of the year', year, '\n')

# extract list of files
files_to_process, _ = utils_proc.get_list(year, folder_database, suffix_tar_correctedxml)
files_to_process.sort()
print('files to process loaded:', files_to_process)

# open dataframe of last names from pickle file
# (there is one file of lastnames per year)
with open(input_lastnames, 'rb') as f:
    df_lastnames = pickle.load(f)

print('dataframe with lastnames loaded')

#%%
# for each file
# TODO !!!! get rid of [66:]
for file_tarpath in files_to_process:
    #print(file_tarpath)
    id_doc = file_tarpath.split('/')[-1][:8]

    # instantiate document object (always from original pdf)
    file_aux = year + '/' + id_doc + '.pdf'
    file_doc = defc.Document(file_aux, folder_database)

    # if document is a discussion
    if (file_doc.check_discussion()) and (id_doc not in ['20032463', '20032952', '20014332']):
        print(id_doc + '\n')
        file_doc.df_lastnames = df_lastnames
        file_doc.annotate_xml()

# Commands to get the compressed version of the file
# (compressed file is around 5 times smaller than uncompressed file)
#data/AB/${year}/05_annotatedxml.tar.gz
utils_proc.compress_tar(output_annotatedxml)




#%%
## to test for one file
#file_tarpath = './1893/20026526_datacorr.xml'
#
#id_doc = file_tarpath.split('/')[-1][:8]
#
## instantiate document object (always from original pdf)
#infile_aux = year + '/' + id_doc + '.pdf'
#file_doc = defc.Document(infile_aux, folder_database)
#
#if (file_doc.check_discussion()) and (id_doc not in ['20032463', '20032952', '20014332']):
    #print(id_doc + '\n')
#
    #file_doc.df_lastnames = df_lastnames
    #file_doc.annotate_xml()


#%%



#id_doc

#len(files_to_process)
